"""
Created on 25 Apr 2023

@author: si
"""


class UnknownEngineType(Exception):
    """
    The engine url either doesn't contain an engine_type or there isn't a registered (subclass of)
    :class:`DataConnector` for the engine_type.

    Engine urls have the format: engine_type://engine_params
    """

    def __init__(self, engine_url):
        """
        @param engine_url: (str)
            The engine_url that contained the unknown engine_type
        """
        super().__init__(f"Unknown engine_type in url: '{engine_url}'")
        self.engine_url = engine_url



import ayeaye


class AbstractModelRunner(ayeaye.PartitionedModel):
    """
    Use a `PartitionedModel` to run some normal :class:`ayeaye.Model`s.

    This is overly simplistic, see the Fossa project (https://github.com/Aye-Aye-Dev/fossa) for how
    is should be done in more complex scenarios.

    This class must be subclassed and given a list of models to run. The models mustn't be
    PartitionedModels (as this could create a resource log jam) and the models mustn't depend on each
    other as this isn't checked.

    The models in the list are started in order but they are run in parallel (with the number of
    parallel tasks depending on the number of CPUs) so any dependencies might not have been satisfied.

    @see :meth:`tests.test_model_partitioned.TestPartitionedModel.test_parallel_models` for an
    example of this running.
    """

    # subclass must implement this. It's a list of tuples (model_cls, kwargs) where kwargs is a
    # dictionary.
    models = None

    def partition_plea(self):
        m_count = len(self.models)
        p = ayeaye.PartitionedModel.PartitionOption(minimum=1, maximum=m_count, optimal=m_count)
        return p

    def partition_slice(self, slice_count):
        target_method = "run_etl_model"  # this is the method subtasks will run
        subtasks = [(target_method, {"model_position": p}) for p in range(len(self.models))]
        return subtasks

    def run_etl_model(self, model_position):
        """
        Run one of the target models in a separate process. Re-direct it's log messages to
        `self.log`.

        @param model_position: (int)
            `self.models` is a list, run the model in this position.
        """

        class LoopBackLogger:
            """
            Redirect log messages from the target model to ModelRunner's log method.
            """

            def __init__(self, log_prefix, log_target):
                self.log_prefix = log_prefix
                self.log_target = log_target

            def write(self, msg):
                self.log_target.log(f"{self.log_prefix} {msg}")

        model_cls, model_kwargs = self.models[model_position]
        model_name = model_cls.__name__

        external_log = LoopBackLogger(log_prefix=model_name, log_target=self)

        self.log(f"Running model {model_name} from position: {model_position}")
        m = model_cls(**model_kwargs)

        m.set_logger(external_log)
        m.log_to_stdout = False  # avoid duplicate messages

        m.go()

    def build(self):
        self.log("Running ModelRunner")


class A(ayeaye.Model):
    def build(self):
        from_context = ayeaye.connector_resolver.resolve("From the build context: {greeting}")
        self.log(f"This is Model A with context: {from_context}")


class B(ayeaye.Model):
    def __init__(self, init_arg, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_arg = init_arg

    def build(self):
        self.log(f"This is Model B with init arg: {self.init_arg}")


class C(ayeaye.Model):
    def build(self):
        self.log("This is Model C")


class ExampleModelRunner(AbstractModelRunner):
    models = [
        (A, {}),
        (B, {"init_arg": "hi model B"}),
        (C, {}),
    ]


if __name__ == "__main__":
    # run it from the command line like this-
    # $ pipenv shell
    # $ python examples/parallel_model_runner/model_runner.py

    build_context = {"greeting": "Hello command line model!"}
    with ayeaye.connector_resolver.context(**build_context):
        m = ExampleModelRunner()
        m.go()



from inspect import isclass

from .base import DataConnector
from .bigquery import BigQueryConnector
from .csv_connector import CsvConnector, TsvConnector
from .elasticsearch_connector import ElasticsearchConnector
from .fake import FakeDataConnector
from .json_connector import JsonConnector
from .kafka_connector import KafkaConnector
from .ndjson_connector import NdjsonConnector
from .parquet_connector import ParquetConnector
from .restful_connector import RestfulConnector, RestfulConnectorConnectionError
from .sqlalchemy_database import SqlAlchemyDatabaseConnector
from .uncooked_connector import UncookedConnector

from ayeaye.ignition import Ignition, EngineUrlCase, EngineUrlStatus
from ayeaye.connectors.engine_type_modifiers import engine_type_modifier_factory
from ayeaye import exception


class ConnectorPluginsRegistry:
    def __init__(self):
        self.registered_connectors = []  # list of classes, not instances - publicly readable
        self.reset()

    def reset(self):
        "set registered connectors to just those that are built in"
        self.registered_connectors = [
            BigQueryConnector,
            CsvConnector,
            FakeDataConnector,
            KafkaConnector,
            ParquetConnector,
            TsvConnector,
            SqlAlchemyDatabaseConnector,
            JsonConnector,
            NdjsonConnector,
            ElasticsearchConnector,
            UncookedConnector,
            RestfulConnector,
        ]

    def register_connector(self, connector_cls):
        """
        Register a non-built in connector.

        e.g.

        class MyConnector(DataConnector):
            engine_type = "my_protocol://"
            ...

        some_data = ayeaye.Connect(engine_url="my_protocol://xyz,abc")

        @param connector_cls (subclass of :class:`DataConnector`):
        """
        if not isclass(connector_cls) or not issubclass(connector_cls, DataConnector):
            msg = "'connector_cls' should be a class (not object) and subclass of DataConnector"
            raise TypeError(msg)

        engine_types = (
            connector_cls.engine_type
            if isinstance(connector_cls.engine_type, list)
            else [connector_cls.engine_type]
        )

        for engine_type in engine_types:
            if not engine_type.endswith("://"):
                msg = "The 'engine_type' class attribute is expected to end with '://'"
                raise ValueError(msg)

            # remove ://
            engine_type = engine_type[0:-3]
            if not engine_type.replace("-", "").replace("_", "").isalnum():
                msg = "The engine_type label can only be alpha-numeric with hyphens and underscores"
                raise ValueError(msg)

        # MAYBE - a mechanism to specify the position/priority of the class
        self.registered_connectors.append(connector_cls)


# global registry
connector_registry = ConnectorPluginsRegistry()


def connector_factory(engine_url):
    """
    @param engine_url (str):
        format is [modifier+[...modifier_n+]]<engine_type>://<engine_specific_parameters>
    @return a subclass of :class:`DataConnector`
    """
    engine_type_separator = "://"
    if isinstance(engine_url, str) and engine_type_separator not in engine_url:
        # The engine type is only available after the context has been resolved. Don't error here
        # if resolution isn't yet possible.
        ignition = Ignition(engine_url)
        try:
            status, e_url = ignition.engine_url_at_state(EngineUrlCase.FULLY_RESOLVED)
            if status == EngineUrlStatus.OK:
                engine_url = e_url
        except ValueError:
            # Full resolve not available yet. It could be possible to use a partially
            # resolved url if this current behaviour isn't good enough.
            pass

    if engine_type_separator in engine_url:
        engine_type = engine_url.split(engine_type_separator, 1)[0] + engine_type_separator

        modifier_labels = None
        if "+" in engine_type:
            # engine type modifiers are being used. e.g. the 'gz' in `gz+ndjson://`
            engine_parts = engine_type.split("+")
            modifier_labels = engine_parts[:-1]
            engine_type = engine_parts[-1]

        for connector_cls in connector_registry.registered_connectors:
            if isinstance(connector_cls.engine_type, list):
                supported_engines = connector_cls.engine_type
            else:
                supported_engines = [connector_cls.engine_type]

            if engine_type in supported_engines:
                if modifier_labels:
                    # find a suitable modifier - it needs to support this connector class
                    # and the modifiers
                    modified_connector_cls = engine_type_modifier_factory(
                        connector_cls=connector_cls,
                        modifier_labels=modifier_labels,
                    )

                    # build a DataConnector like class that can be treated just like a normal
                    # data connector.
                    dynamic_dc_cls = modified_connector_cls.apply(
                        connector_cls=connector_cls,
                        modifier_labels=modifier_labels,
                    )

                    return dynamic_dc_cls

                return connector_cls

    raise exception.UnknownEngineType(engine_url=engine_url)



from inspect import isclass

from .abstract_modifier import AbstractEngineTypeModifier
from .smart_open_modifier import SmartOpenModifier


def engine_type_modifier_factory(connector_cls, modifier_labels):
    """
    Return a single modifier that supports all of the `modifier_labels` and can work with the
    `connector_cls`.

    Behind the scenes this could use composition, dynamic inheritance etc. For now, it's
    simple - the returned modifier simultaneously supports all of the specified modifier_labels.

    @param connector_cls: (subclass of :class:`DataConnector`)
        The target data connector for the modifier.

    @param modifier_labels (list of str):
        Named properties that must all be supported by the returned modifier.
        e.g. ['s3', 'gz']

    @return (subclass of :class:`AbstractEngineTypeModifier`)
    """

    for modifier_cls in connector_modifiers_registry.registered_modifiers:
        if modifier_cls.provides_support(connector_cls, modifier_labels):
            return modifier_cls

    mod_labels = "+".join(modifier_labels)
    msg = (
        f"The combination of {connector_cls.__name__} and {mod_labels} engine type modifiers "
        "isn't supported."
    )
    raise NotImplementedError(msg)


class ConnectorModifiersPluginsRegistry:
    """
    A modifier adds capabilities to a connector, for example, transparent compression.
    """

    def __init__(self):
        self.registered_modifiers = []  # list of classes, not instances - publicly readable
        self.reset()

    def reset(self):
        "set registered connectors to just those that are built in"
        self.registered_modifiers = [
            SmartOpenModifier,
        ]

    def register_connector(self, modifier_cls):
        """
        Add a private modifier to the dataset connection discovery process.
        @param modifier_cls (subclass of :class:`AbstractEngineTypeModifier`):
        """
        if not isclass(modifier_cls) or not issubclass(modifier_cls, AbstractEngineTypeModifier):
            msg = "'connector_cls' should be a class (not object) and subclass of AbstractEngineTypeModifier"
            raise TypeError(msg)

        # MAYBE - a mechanism to specify the position/priority of the class
        self.registered_connectors.append(modifier_cls)


# global registry
connector_modifiers_registry = ConnectorModifiersPluginsRegistry()



from datetime import datetime, timezone
import os

try:
    from smart_open import open as smart_open
    import boto3
except ModuleNotFoundError:
    pass

from ayeaye.connectors.base import FileBasedConnector, FilesystemEnginePattern
from ayeaye.connectors.engine_type_modifiers.abstract_modifier import AbstractEngineTypeModifier
from ayeaye.connectors.engine_type_modifiers.utils import s3_pattern_match
from ayeaye.ignition import EngineUrlCase, EngineUrlStatus


class SmartOpenEnginePattern(FilesystemEnginePattern):
    """
    Use wildcards to pattern match files and directories. @see :class:`AbstractExpandEnginePattern`.

    The class adds S3 pattern expansion to the filesystem expansion provided by
    :class:`FilesystemEnginePattern`.
    """

    def expand_pattern(self):
        if not self.data_connector._s3_resource:
            # Not working with an S3 hosted file so must be using a local filesystem
            # so use the parent's method.
            return super().expand_pattern()

        status, e_url = self.data_connector.ignition.engine_url_at_state(
            EngineUrlCase.FULLY_RESOLVED
        )
        assert status == EngineUrlStatus.OK, "Super class should have caught this."

        # strip engine type
        # for now, they all need to have the same engine_type. Maybe engine_url starts
        # with `://` for auto detect based on file name.
        engine_type, engine_path_pattern = e_url.split("://", 1)

        if "?" in engine_path_pattern:
            raise NotImplementedError("TODO: Sorry, '?' isn't yet supported.")

        if "/" not in engine_path_pattern:
            raise ValueError(f"Bucket name not found in: {e_url}")

        bucket_name, obj_key_pattern = engine_path_pattern.split("/", 1)

        # the S3 API can only filter by the start of the file/object name. The rest is done with
        # a regular expression.
        prefix, matcher = s3_pattern_match(obj_key_pattern)

        s3_client = self.data_connector._s3_client
        continuation_token = None
        s3_kwargs = {"Bucket": bucket_name, "Prefix": prefix}
        engine_url = []
        while True:
            response = s3_client.list_objects_v2(**s3_kwargs)
            content = response.get("Contents", [])
            for c in content:
                engine_file = c["Key"]
                if matcher(engine_file):
                    engine_url.append(f"{engine_type}://{engine_file}")

            continuation_token = response.get("NextContinuationToken", None)
            if continuation_token:
                # 2nd page onwards
                s3_kwargs["ContinuationToken"] = continuation_token
            else:
                # end of pages
                break

        return engine_url


class SmartOpenModifier(AbstractEngineTypeModifier):
    """
    Use a sub-set of smart-open (https://pypi.org/project/smart-open/) to-
    - open s3 blobs
    - transparently compress and de-compress
    - expand file patterns using wild cards

    for all sub-classes of :class:`FilesystemConnector`.

    Implementation : This class is used to override methods in :class:`FilesystemConnector`.
    """

    engine_pattern_expander_cls = SmartOpenEnginePattern

    def __init__(self):
        # lazy load variables
        self.__remote_file_attribs = None
        self.__s3_client = None

    @staticmethod
    def provides_support(connector_cls, modifier_labels):
        supported_labels = set(["gz", "s3"])
        proposed = set(modifier_labels)

        supported = proposed.issubset(supported_labels) and issubclass(
            connector_cls, FileBasedConnector
        )
        return supported

    @property
    def _s3_resource(self):
        return "s3" in self.requested_modifier_labels

    @property
    def _s3_client(self):
        if self.__s3_client is None:
            self.__s3_client = boto3.client("s3")

        return self.__s3_client

    @property
    def _remote_file_attribs(self):
        """
        @return: dict or None (if file doesn't exist)
            dict-
            'file_size' : int or None (if not available)
        """
        assert self._s3_resource, "should only be used in S3 mode"

        if self.__remote_file_attribs is None:
            bucket_name, obj_key = self.file_path.split("/", 1)

            try:
                r = self._s3_client.get_object_attributes(
                    Bucket=bucket_name,
                    Key=obj_key,
                    ObjectAttributes=["ObjectSize"],
                )
            except self._s3_client.exceptions.NoSuchKey:
                return

            self.__remote_file_attribs = {"file_size": r.get("ObjectSize")}

        return self.__remote_file_attribs

    def _open(self, *args, **kwargs):
        """
        Overrides :meth:`FilesystemConnector.connect` with one using Smart Open's open.
        """
        smart_open_kwargs = {}
        if "gz" in self.requested_modifier_labels:
            smart_open_kwargs["compression"] = ".gz"
        else:
            smart_open_kwargs["compression"] = "disable"

        if self._s3_resource:
            # first arg is always the file path. Pre-fix this for smart open
            args = tuple(["s3://" + args[0]] + list(args[1:]))

        return smart_open(*args, **kwargs, **smart_open_kwargs)

    @property
    def datasource_exists(self):
        """
        Returns:
            (bool) if the datasource referred to in self.engine_url exists.
        """
        if self._s3_resource:
            return self._remote_file_attribs is not None

        return os.path.exists(self.file_path)

    def _get_file_size(self):
        """
        @return: int or None if not available
        """
        if self.datasource_exists:
            if self._s3_resource:
                return self._remote_file_attribs.get("file_size")

            return os.stat(self.file_path).st_size
        return None

    @property
    def last_modified(self):
        """
        Returns:
            (UTC `datetime.datetime`) of file, or None if file does not exist
        """
        if not self.datasource_exists:
            return None

        if self._s3_resource:
            raise NotImplementedError("TODO: S3, get from :meth:`_remote_file_attribs`")

        timestamp = os.path.getmtime(self.file_path)
        last_modified = datetime.utcfromtimestamp(timestamp).replace(tzinfo=timezone.utc)
        return last_modified

    def auto_create_directory(self):
        if self._s3_resource:
            # I'm pretty sure you don't need to do this with s3/gcs
            return

        self._auto_create_directory()



"""
This link between the runtime 'execution' environment and a worker process running a model.

Created on 31 Jan 2022

@author: si
"""
import os


class RuntimeKnowledge:
    """
    This link between the runtime 'execution' environment and a process running a model.

    This module will develop to make it possible for a model to introspect and communicate with
    it's environment. The process holding an instance of this class could be a worker or
    the main process that partitions the task into sub-tasks for the worker.

    Available attributes-

    * worker_id (int) - unique integer assigned in ascending order to workers as they start
    * total_workers (int) - Number of workers created in worker group. Is `None` if variable number
                    of workers are being used.
    """

    def __init__(self):
        self.worker_id = None
        self.total_workers = None

        # Number of concurrent tasks per CPU
        self.cpu_task_ratio = 2

        # can be set to an absolute limit
        # e.g.
        #   model_instance.runtime.max_concurrent_tasks = 1
        # or by default returns number of CPUs * self.cpu_task ratio
        self._max_concurrent_tasks = None

    @property
    def max_concurrent_tasks(self):
        if self._max_concurrent_tasks is not None:
            # user has set an absolute value
            return self._max_concurrent_tasks

        return os.cpu_count() * self.cpu_task_ratio

    @max_concurrent_tasks.setter
    def max_concurrent_tasks(self, max_tasks):
        """
        Override the system default and set a maximum absolute number of concurrent tasks.

        @param max_tasks: (int)
        """
        self._max_concurrent_tasks = max_tasks



import re


def s3_pattern_match(key_pattern):
    """
    The S3 API can only filter by the start of the file/object name. The rest is done with
    a regular expression.

    @param key_pattern: (str)
        A key with a wildcard in it. e.g. 'my_directory/sub_directory/*.csv'

    @return: prefix, re_matcher (str, func)
        func takes a single argument, the full path and returns True if it matches the pattern.
    """
    assert "*" in key_pattern

    # regex safety. This does restrict some legitimate object paths hence the type of exception
    # raised.
    reg_ex_reserved = "?\()!^$+{}[]|<"
    for char in reg_ex_reserved:
        if char in key_pattern:
            msg = f"Sorry, object pattern not supported. It can't have '{reg_ex_reserved}' in it."
            raise NotImplementedError(msg)

    prefix = key_pattern.split("*", 1)[0]

    # make it into a regex
    expression = key_pattern.replace(".", "\.").replace("*", ".*")
    pattern = re.compile(expression)

    def check_file_path(full_path):
        return pattern.fullmatch(full_path) is not None

    return prefix, check_file_path



class AbstractEngineTypeModifier:
    """
    engine_urls can be prefixed with modifier that slightly changes how the underlying
    DataConnector works.

    For example-
        engine_url="gz+ndjson:///data/myfile.ndjson.gz"

    Would transparently compress/decompress on the fly but is otherwise using the
    :class:`ayeaye.Connectors.NdjsonConnector` connector.

    Modifier classes are subclasses of this `AbstractEngineTypeModifier` class and their :meth:`apply` is
    applied to the target `DataConnector` in the same way that python decorators work behind the
    scenes-

    e.g.

    >>> my_connector = NdjsonConnector(**some_kwargs)
    >>> a_modifer = ExampleCloudStorageModifier()
    >>> my_connector = a_modifer.apply(connector=my_connector)
    """

    @staticmethod
    def provides_support(connector_cls, modifier_labels):
        """
        Does this (i.e. `self`) engine type modifier support all of the `modifier_labels` for the
        `connector_cls`?

        @param connector_cls: (subclass of :class:`DataConnector`)
            The target data connector for the modifier.

        @param modifier_labels (list of str):
            Named properties that must all be supported by the returned modifier. It's possible
            that the order of these matters, hence a list.
            e.g. ['s3', 'gz']

        @return bool
        """
        raise NotImplementedError("Must be implemented by subclasses")

    @classmethod
    def apply(cls, connector_cls, modifier_labels):
        """
        Build a `DataConnector` like object that can take the place of a true :class:`DataConnector`
        instance.

        @param connector_cls: (subclass of :class:`DataConnector`)
            The target data connector for the modifier.

        @param modifier_labels (list of str):
            Named properties that must all be supported by the returned modifier. It's possible
            that the order of these matters, hence a list.
            e.g. ['s3', 'gz']


        @return: a :class:`DataConnector` like object.
        """

        # create engine_type for dynamic class
        mod_labels = "+".join(modifier_labels)
        supported_engines = (
            connector_cls.engine_type
            if isinstance(connector_cls.engine_type, list)
            else [connector_cls.engine_type]
        )
        dyn_engine_type = [mod_labels + "+" + engine for engine in supported_engines]

        class DynamicConnector(cls, connector_cls):
            requested_modifier_labels = modifier_labels
            engine_type = dyn_engine_type

            def __init__(self, *args, **kwargs):
                cls.__init__(self)
                connector_cls.__init__(self, *args, **kwargs)

        return DynamicConnector



class MultiConnectorNewDataset:
    """
    Multi-connectors are often used to produce a number of similar files. This helper class makes
    an additional method on a :class:`MultiConnector` connection to make it easy to create files
    using a templated engine_url.

    This is done using the `method_overlay` argument to the :class:`DataConnector` superclass. It
    can be used like this-

    # (as a class variable)
    output_file_template = "csv://{product_group}_{product_name}_parts_list.csv"
    components_doc = ayeaye.Connect(
        engine_url=[],
        method_overlay=(MultiConnectorNewDataset(template=output_file_template), "new_dataset"),
        access=ayeaye.AccessMode.WRITE,
    )
    ...
    # then use it like this ...
    components = self.components_doc.new_dataset(product_group="machinery", product_name="digger")

    # components is now a new CSV data connector, add something to it
    components.add({"name": "spring", "product_code": "ab123"})

    Also, see unittest :class:`TestConnectHelper`.
    """

    def __init__(self, template):
        """
        @param template: (str)
            template for an engine_url
        """
        self.template = template

    def __call__(self, parent_connector, *args, **kwargs):
        """
        @param kwargs: dict with strings for both key and value
            These will be substituted into `self.template`
        """
        resolved_template = self.template
        for k, v in kwargs.items():
            template_field = "{" + k + "}"
            if template_field in self.template:
                # can't use .format as not all template fields are being replaced
                resolved_template = resolved_template.replace(template_field, v)

        new_dataset_connection = parent_connector.add_engine_url(resolved_template)
        return new_dataset_connection
